import "dart:collection" show IterableMixin;
import "package:meta/meta.dart";
import "package:poker/poker.dart";

@immutable
class CardSet with IterableMixin<Card> {
  /// An empty CardSet.
  static const CardSet empty = CardSet.fromInt(0);

  /// A full-deck CardSet.
  static const CardSet full = CardSet.fromInt(4503599627370495);

  /// Creates a CardSet from a Iterable<Card>.
  CardSet(Iterable<Card> cards)
      : _value = cards.fold<int>(0, (v, card) => v + card.toInt());

  CardSet.single(Card card) : _value = card.toInt();

  /// Creates a CardSet from a hash int value.
  const CardSet.fromInt(int value) : _value = value;

  factory CardSet.fromString(String string) {
    CardSet cards = CardSet.empty;

    for (int i = 0; i < string.length; i += 2) {
      try {
        cards = cards.union(
          CardSet.single(Card.fromString(string.substring(i, i + 2))),
        );
      } catch (error) {
        throw CardSetParseFailureException(string);
      }
    }

    return cards;
  }

  final int _value;

  @override
  Iterator<Card> get iterator => CardSetIterator(this);

  @override
  bool contains(Object? other) =>
      other is Card &&
      intersection(CardSet.single(other)) == CardSet.single(other);

  /// Whether this CardSet contains all the cards in the given CardSet or not.
  bool containsAll(Object? other) =>
      other is CardSet && intersection(other) == other;

  /// Returns an union CardSet of this and the given CardSet.
  CardSet union(CardSet other) => CardSet.fromInt(toInt() | other.toInt());

  /// Returns an difference CardSet of this and the given CardSet.
  CardSet difference(CardSet other) =>
      CardSet.fromInt(toInt() & ~other.toInt());

  /// Returns an intersection CardSet of this and the given CardSet.
  CardSet intersection(CardSet other) =>
      CardSet.fromInt(toInt() & other.toInt());

  /// Returns the hash int value of this CardSet.
  int toInt() => _value;

  @override
  int get hashCode => toInt();

  @override
  operator ==(Object other) => other is CardSet && toInt() == other.toInt();
}

class CardSetIterator implements Iterator<Card> {
  CardSetIterator(this.original)
      : _remaining = original.toInt(),
        _current = null;

  final CardSet original;

  int _remaining;

  Card? _current;

  @override
  Card get current => _current!;

  @override
  bool moveNext() {
    if (_remaining > 0) {
      _current = Card.fromInt(_remaining & -_remaining);
      _remaining = _remaining & (_remaining - 1);

      return true;
    }

    return false;
  }
}

class CardSetParseFailureException implements Exception {
  CardSetParseFailureException(this.value);

  final String value;

  @override
  String toString() {
    return "CardSetParseFailureException: $value is not a valid string.";
  }
}
