import 'dart:collection';

import "package:meta/meta.dart";
import "package:poker/poker.dart";

@immutable
class HandRange with IterableMixin<CardPair> {
  HandRange(Set<CardPairConvertable> components) : components = components;

  factory HandRange.fromString(String value) {
    final components = Set<CardPairConvertable>();
    final matches = RegExp(
            r"[AKQJT98765432][shdc]?[AKQJT98765432](\+|[so]\+?|[shdc])?(-[AKQJT98765432][AKQJT98765432][so]?)?")
        .allMatches(value);

    for (final match in matches) {
      final token = match.group(0)!;

      // parse range parts like AKs-ATs
      if (token.length == 7) {
        final high = rankFromChar(token[0]);
        final kickerTop = rankFromChar(token[1]);
        final kickerBottom = rankFromChar(token[5]);

        for (int i = _ranks.indexOf(kickerTop);
            i <= _ranks.indexOf(kickerBottom);
            ++i) {
          if (token[2] == "s") {
            components.add(RankPair.suited(high: high, kicker: _ranks[i]));
          } else {
            components.add(RankPair.ofsuit(high: high, kicker: _ranks[i]));
          }
        }

        continue;
      }

      // parse range parts like QQ-55
      if (token.length == 5 && token[0] == token[1] && token[3] == token[4]) {
        final top = rankFromChar(token[0]);
        final bottom = rankFromChar(token[3]);

        for (int i = _ranks.indexOf(top); i <= _ranks.indexOf(bottom); ++i) {
          components.add(RankPair.ofsuit(high: _ranks[i], kicker: _ranks[i]));
        }

        continue;
      }

      // parse pocket range parts like 55+
      if (token.length == 3 && token[2] == "+" && token[0] == token[1]) {
        for (int i = 0; i <= _ranks.indexOf(rankFromChar(token[0])); ++i) {
          components.add(RankPair.ofsuit(high: _ranks[i], kicker: _ranks[i]));
        }

        continue;
      }

      // parse range parts like AJs+, AJ+
      if ((token.length == 3 && token[2] == "+" && token[0] != token[1]) ||
          (token.length == 4 && token[3] == "+")) {
        final high = rankFromChar(token[0]);
        final kicker = rankFromChar(token[1]);

        for (int i = _ranks.indexOf(high) + 1;
            i <= _ranks.indexOf(kicker);
            ++i) {
          if (token[2] == "s") {
            components.add(RankPair.suited(high: high, kicker: kicker));
          } else {
            components.add(RankPair.ofsuit(high: high, kicker: kicker));
          }
        }

        continue;
      }

      // example: AhKc
      if (token.length == 4 && token[3] != "+") {
        components.add(CardPair(
          Card.fromString(token.substring(0, 2)),
          Card.fromString(token.substring(2, 4)),
        ));

        continue;
      }

      final high = rankFromChar(token[0]);
      final kicker = rankFromChar(token[1]);

      // example: AJ, AJo
      if (token[2] == "s") {
        components.add(RankPair.suited(high: high, kicker: kicker));
      } else {
        if (high == kicker) {
          components.add(RankPair.ofsuit(high: high, kicker: kicker));
        } else {
          components.add(RankPair.ofsuit(high: high, kicker: kicker));
        }
      }
    }

    return HandRange(components);
  }

  final Set<CardPairConvertable> components;

  Iterator<CardPair> get iterator => components
      .fold<Set<CardPair>>(
        LinkedHashSet(),
        (set, component) => set..addAll(component.toCardPairs()),
      )
      .iterator;

  // Set<CardPair> get onlyCardPairs => components.whereType<CardPair>().toSet();

  // bool get hasCardPair => onlyCardPairs.length >= 1;

  // Set<RankPair> get onlyRankPairs => components.whereType<RankPair>().toSet();

  // bool get hasRankPair => onlyRankPairs.length >= 1;

  @override
  bool get isEmpty => components.length == 0;

  @override
  String toString() {
    String result = "";

    int start = -1;
    for (int i = 0; i < _ranks.length; ++i) {
      if (components
          .contains(RankPair.ofsuit(high: _ranks[i], kicker: _ranks[i]))) {
        if (start == -1) {
          start = i;
        }
      }

      if (start != -1 &&
          (i == _ranks.length - 1 ||
              !components.contains(RankPair.ofsuit(
                  high: _ranks[i + 1], kicker: _ranks[i + 1])))) {
        if (start == i) {
          result += "${_ranks[i].char}${_ranks[i].char}";
        } else if (start == 0) {
          result += "${_ranks[i].char}${_ranks[i].char}+";
        } else {
          result +=
              "${_ranks[start].char}${_ranks[start].char}-${_ranks[i].char}${_ranks[i].char}";
        }

        start = -1;
      }
    }

    for (int h = 0; h < _ranks.length - 1; ++h) {
      int suitedStart = -1;
      int ofsuitStart = -1;

      for (int k = h + 1; k < _ranks.length; ++k) {
        if (components
            .contains(RankPair.suited(high: _ranks[h], kicker: _ranks[k]))) {
          if (suitedStart == -1) {
            suitedStart = k;
          }
        }

        if (suitedStart != -1 &&
            (k == _ranks.length - 1 ||
                !components.contains(
                    RankPair.suited(high: _ranks[h], kicker: _ranks[k + 1])))) {
          if (suitedStart == k) {
            result += "${_ranks[h].char}${_ranks[k].char}s";
          } else if (suitedStart == h + 1) {
            result += "${_ranks[h].char}${_ranks[k].char}s+";
          } else {
            result +=
                "${_ranks[h].char}${_ranks[suitedStart].char}s-${_ranks[h].char}${_ranks[k].char}s";
          }

          suitedStart = -1;
        }
      }

      for (int k = h + 1; k < _ranks.length; ++k) {
        if (components
            .contains(RankPair.ofsuit(high: _ranks[h], kicker: _ranks[k]))) {
          if (ofsuitStart == -1) {
            ofsuitStart = k;
          }
        }

        if (ofsuitStart != -1 &&
            (k == _ranks.length - 1 ||
                !components.contains(
                    RankPair.ofsuit(high: _ranks[h], kicker: _ranks[k + 1])))) {
          if (ofsuitStart == k) {
            result += "${_ranks[h].char}${_ranks[k].char}o";
          } else if (ofsuitStart == h + 1) {
            result += "${_ranks[h].char}${_ranks[k].char}o+";
          } else {
            result +=
                "${_ranks[h].char}${_ranks[ofsuitStart].char}o-${_ranks[h].char}${_ranks[k].char}o";
          }

          ofsuitStart = -1;
        }
      }
    }

    for (final cardPair in components.whereType<CardPair>()) {
      result +=
          "${cardPair._cardA.toShortString()}${cardPair._cardB.toShortString()}";
    }

    return "HandRange<$result>";
  }
}

mixin CardPairConvertable implements Comparable {
  Set<CardPair> toCardPairs();

  @override
  int compareTo(dynamic other) {
    if (other is! CardPairConvertable) {
      return 0;
    }

    if (runtimeType != other.runtimeType) {
      return this is RankPair ? -1 : 1;
    }

    if (this is RankPair && other is RankPair) {
      final self = this as RankPair;

      if (self.high != other.high) {
        return self.high.powerIndex - other.high.powerIndex;
      }

      if (self.kicker != other.kicker) {
        return self.kicker.powerIndex - other.kicker.powerIndex;
      }

      if (self.isSuited && !other.isSuited) {
        return -1;
      }

      if (!self.isSuited && other.isSuited) {
        return 1;
      }

      return 0;
    }

    if (this is CardPair && other is CardPair) {
      final self = this as CardPair;

      if (self._cardA.rank != other._cardA.rank) {
        return self._cardA.rank.powerIndex - other._cardA.rank.powerIndex;
      }

      if (self._cardB.rank != other._cardB.rank) {
        return self._cardB.rank.powerIndex - other._cardB.rank.powerIndex;
      }

      if (self._cardA.suit != other._cardA.suit) {
        return self._cardA.suit.index - other._cardA.suit.index;
      }

      return self._cardB.suit.index - other._cardB.suit.index;
    }

    return 0;
  }
}

class CardPair
    with CardPairConvertable, IterableMixin<Card>
    implements CardSet {
  CardPair(Card cardA, Card cardB)
      : _cardA = cardA.rank.powerIndex <= cardB.rank.powerIndex ? cardA : cardB,
        _cardB = cardA.rank.powerIndex <= cardB.rank.powerIndex ? cardB : cardA;

  final Card _cardA;

  final Card _cardB;

  @override
  Set<CardPair> toCardPairs() => {this};

  @override
  int toInt() => CardSet({_cardA, _cardB}).toInt();

  @override
  Iterator<Card> get iterator => CardSet({_cardA, _cardB}).iterator;

  @override
  bool containsAll(Object? other) =>
      CardSet({_cardA, _cardB}).containsAll(other);

  @override
  CardSet union(CardSet other) => CardSet({_cardA, _cardB}).union(other);

  @override
  CardSet intersection(CardSet other) =>
      CardSet({_cardA, _cardB}).intersection(other);

  @override
  CardSet difference(CardSet other) =>
      CardSet({_cardA, _cardB}).difference(other);

  @override
  String toString() =>
      "CardPair<${_cardA.toShortString()}${_cardB.toShortString()}>";

  operator ==(Object other) =>
      other is CardPair && _cardA == other._cardA && _cardB == other._cardB;

  @override
  int get hashCode {
    int result = 17;

    result = 37 * result + _cardA.hashCode;
    result = 37 * result + _cardB.hashCode;

    return result;
  }
}

@immutable
class RankPair with CardPairConvertable {
  const RankPair.suited({required this.high, required this.kicker})
      : assert(high != kicker),
        isSuited = true;

  const RankPair.ofsuit({required this.high, required this.kicker})
      : isSuited = false;

  final Rank high;

  final Rank kicker;

  final bool isSuited;

  bool get isPocketPair => high == kicker;

  @override
  Set<CardPair> toCardPairs() => high == kicker
      ? {
          CardPair(
            Card(rank: high, suit: Suit.spade),
            Card(rank: kicker, suit: Suit.heart),
          ),
          CardPair(
            Card(rank: high, suit: Suit.spade),
            Card(rank: kicker, suit: Suit.diamond),
          ),
          CardPair(
            Card(rank: high, suit: Suit.spade),
            Card(rank: kicker, suit: Suit.club),
          ),
          CardPair(
            Card(rank: high, suit: Suit.heart),
            Card(rank: kicker, suit: Suit.diamond),
          ),
          CardPair(
            Card(rank: high, suit: Suit.heart),
            Card(rank: kicker, suit: Suit.club),
          ),
          CardPair(
            Card(rank: high, suit: Suit.diamond),
            Card(rank: kicker, suit: Suit.club),
          ),
        }
      : isSuited
          ? {
              CardPair(
                Card(rank: high, suit: Suit.spade),
                Card(rank: kicker, suit: Suit.spade),
              ),
              CardPair(
                Card(rank: high, suit: Suit.heart),
                Card(rank: kicker, suit: Suit.heart),
              ),
              CardPair(
                Card(rank: high, suit: Suit.diamond),
                Card(rank: kicker, suit: Suit.diamond),
              ),
              CardPair(
                Card(rank: high, suit: Suit.club),
                Card(rank: kicker, suit: Suit.club),
              ),
            }
          : {
              CardPair(
                Card(rank: high, suit: Suit.spade),
                Card(rank: kicker, suit: Suit.heart),
              ),
              CardPair(
                Card(rank: high, suit: Suit.spade),
                Card(rank: kicker, suit: Suit.diamond),
              ),
              CardPair(
                Card(rank: high, suit: Suit.spade),
                Card(rank: kicker, suit: Suit.club),
              ),
              CardPair(
                Card(rank: high, suit: Suit.heart),
                Card(rank: kicker, suit: Suit.spade),
              ),
              CardPair(
                Card(rank: high, suit: Suit.heart),
                Card(rank: kicker, suit: Suit.diamond),
              ),
              CardPair(
                Card(rank: high, suit: Suit.heart),
                Card(rank: kicker, suit: Suit.club),
              ),
              CardPair(
                Card(rank: high, suit: Suit.diamond),
                Card(rank: kicker, suit: Suit.spade),
              ),
              CardPair(
                Card(rank: high, suit: Suit.diamond),
                Card(rank: kicker, suit: Suit.heart),
              ),
              CardPair(
                Card(rank: high, suit: Suit.diamond),
                Card(rank: kicker, suit: Suit.club),
              ),
              CardPair(
                Card(rank: high, suit: Suit.club),
                Card(rank: kicker, suit: Suit.spade),
              ),
              CardPair(
                Card(rank: high, suit: Suit.club),
                Card(rank: kicker, suit: Suit.heart),
              ),
              CardPair(
                Card(rank: high, suit: Suit.club),
                Card(rank: kicker, suit: Suit.diamond),
              ),
            };

  @override
  String toString() =>
      "${high.char}${kicker.char}${isPocketPair ? "" : isSuited ? "s" : "o"}";

  @override
  int get hashCode {
    int result = 17;

    result = 37 * result + high.hashCode;
    result = 37 * result + kicker.hashCode;
    result = 37 * result + isSuited.hashCode;

    return result;
  }

  @override
  operator ==(Object other) =>
      other is RankPair &&
      other.high == high &&
      other.kicker == kicker &&
      other.isSuited == isSuited;
}

const List<Rank> _ranks = [
  Rank.ace,
  Rank.king,
  Rank.queen,
  Rank.jack,
  Rank.ten,
  Rank.nine,
  Rank.eight,
  Rank.seven,
  Rank.six,
  Rank.five,
  Rank.four,
  Rank.trey,
  Rank.deuce,
];
