import "package:poker/poker.dart";
import "package:test/test.dart";

void main() {
  test("CardSet.empty has a 52-bit hash int value that its every bit is 0", () {
    expect(
        CardSet.empty.toInt(),
        equals(int.parse(
          "0000000000000000000000000000000000000000000000000000",
          radix: 2,
        )));
  });

  test("CardSet.full has a 52-bit hash int value that its every bit is 1", () {
    expect(
        CardSet.full.toInt(),
        equals(int.parse(
          "1111111111111111111111111111111111111111111111111111",
          radix: 2,
        )));
  });

  group("CardSet.fromString()", () {
    test(
        "CardSet.fromString(\"AsKcQhJd\") returns a CardSet<{ As, Qh, Jd, Kh }>",
        () {
      expect(
        CardSet.fromString("AsKcQhJd"),
        equals(CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        })),
      );
    });

    test("CardSet.fromString(\"\") returns an empty CardSet", () {
      expect(
        CardSet.fromString(""),
        equals(CardSet.empty),
      );
    });

    test("CardSet.fromString(\"AsKC\") throws a CardSetParseFailureException",
        () {
      expect(
        () => CardSet.fromString("AsKC"),
        throwsA(isA<CardSetParseFailureException>()),
      );
    });

    test("CardSet.fromString(\"AsKcqh\") throws a CardSetParseFailureException",
        () {
      expect(
        () => CardSet.fromString("AsKcqh"),
        throwsA(isA<CardSetParseFailureException>()),
      );
    });

    test("CardSet.fromString(\"AKs\") throws a CardSetParseFailureException",
        () {
      expect(
        () => CardSet.fromString("AKs"),
        throwsA(isA<CardSetParseFailureException>()),
      );
    });

    test("CardSet.fromString(\"44\") throws a CardSetParseFailureException",
        () {
      expect(
        () => CardSet.fromString("44"),
        throwsA(isA<CardSetParseFailureException>()),
      );
    });

    test("CardSet.fromString(\"As+\") throws a CardSetParseFailureException",
        () {
      expect(
        () => CardSet.fromString("As+"),
        throwsA(isA<CardSetParseFailureException>()),
      );
    });
  });

  group("CardSet#length", () {
    test("CardSet<{ As, Qh, Jd, Kc }>.length returns 4", () {
      expect(
        CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        }).length,
        equals(4),
      );
    });

    test("CardSet<{ As }>.length returns 1", () {
      expect(
        CardSet({Card(rank: Rank.ace, suit: Suit.spade)}).length,
        equals(1),
      );
    });

    test("CardSet.empty.length returns 0", () {
      expect(CardSet.empty.length, equals(0));
    });

    test("CardSet.full.length returns 52", () {
      expect(CardSet.full.length, equals(52));
    });
  });

  group("CardSet#containsAll()", () {
    test("CardSet<AsQhJdKh>.containsAll(CardSet<AsJd>) returns true", () {
      expect(
        CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        }).containsAll(CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.jack, suit: Suit.diamond)
        })),
        isTrue,
      );
    });

    test("CardSet<2c>.containsAll(CardSet.empty) returns true", () {
      expect(
        CardSet({Card(rank: Rank.deuce, suit: Suit.club)})
            .containsAll(CardSet.empty),
        isTrue,
      );
    });

    test("CardSet.empty.containsAll(CardSet<2c>) returns false", () {
      expect(
        CardSet.empty
            .containsAll(CardSet({Card(rank: Rank.deuce, suit: Suit.club)})),
        isFalse,
      );
    });

    test("CardSet.empty.containsAll(CardSet.empty) returns true", () {
      expect(CardSet.empty.containsAll(CardSet.empty), isTrue);
    });

    test("CardSet.full.containsAll(CardSet.full) returns true", () {
      expect(CardSet.full.containsAll(CardSet.full), isTrue);
    });
  });

  group("CardSet#union()", () {
    test("CardSet<AsJd>.union(CardSet<QhKh>) returns CardSet<AsQhJdKh>", () {
      expect(
        CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.jack, suit: Suit.diamond),
        }).union(CardSet({
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.king, suit: Suit.club),
        })),
        equals(CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        })),
      );
    });

    test("CardSet<As>.union(CardSet<As>) returns CardSet<As>", () {
      expect(
        CardSet({Card(rank: Rank.ace, suit: Suit.spade)})
            .union(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
        equals(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
      );
    });

    test("CardSet<As>.union(CardSet.empty) returns CardSet<As>", () {
      expect(
        CardSet({Card(rank: Rank.ace, suit: Suit.spade)}).union(CardSet.empty),
        equals(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
      );
    });

    test("CardSet.empty.union(CardSet<As>) returns CardSet<As>", () {
      expect(
        CardSet.empty.union(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
        equals(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
      );
    });

    test("CardSet.empty.union(CardSet.empty) returns CardSet.empty", () {
      expect(CardSet.empty.union(CardSet.empty), equals(CardSet.empty));
    });

    test("CardSet.full.union(CardSet.full) returns CardSet.full", () {
      expect(CardSet.full.union(CardSet.full), equals(CardSet.full));
    });
  });

  group("CardSet#difference()", () {
    test("CardSet<AsJd>.difference(CardSet<QhKh>) returns CardSet<AsQhJdKh>",
        () {
      expect(
        CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        }).difference(CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.jack, suit: Suit.diamond),
        })),
        equals(CardSet({
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.king, suit: Suit.club),
        })),
      );
    });

    test("CardSet<As>.difference(CardSet.empty) returns CardSet<As>", () {
      expect(
        CardSet({Card(rank: Rank.ace, suit: Suit.spade)})
            .difference(CardSet.empty),
        equals(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
      );
    });

    test("CardSet.empty.difference(CardSet<As>) returns CardSet.empty", () {
      expect(
        CardSet.empty
            .difference(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
        equals(CardSet.empty),
      );
    });

    test("CardSet.empty.difference(CardSet.empty) returns CardSet.empty", () {
      expect(
        CardSet.empty.difference(CardSet.empty),
        equals(CardSet.empty),
      );
    });

    test("CardSet.full.difference(CardSet.full) returns CardSet.empty", () {
      expect(
        CardSet.full.difference(CardSet.full),
        equals(CardSet.empty),
      );
    });
  });

  group("CardSet#intersection()", () {
    test("CardSet<AsQhJdKc>.intersection(CardSet<QhKh>) returns CardSet<QhKh>",
        () {
      expect(
        CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        }).intersection(CardSet({
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.king, suit: Suit.club),
        })),
        equals(CardSet({
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.king, suit: Suit.club),
        })),
      );
    });

    test("CardSet<As>.intersection(CardSet.empty) returns CardSet.empty", () {
      expect(
        CardSet({Card(rank: Rank.ace, suit: Suit.spade)})
            .intersection(CardSet.empty),
        equals(CardSet.empty),
      );
    });

    test("CardSet.empty.intersection(CardSet<As>) returns CardSet.empty", () {
      expect(
        CardSet.empty
            .intersection(CardSet({Card(rank: Rank.ace, suit: Suit.spade)})),
        equals(CardSet.empty),
      );
    });

    test("CardSet.empty.intersection(CardSet.empty) returns CardSet.empty", () {
      expect(
        CardSet.empty.intersection(CardSet.empty),
        equals(CardSet.empty),
      );
    });

    test("CardSet.full.intersection(CardSet.full) returns CardSet.full", () {
      expect(
        CardSet.full.intersection(CardSet.full),
        equals(CardSet.full),
      );
    });
  });

  group("CardSet#iterator", () {
    test(
        "List.from(CardSet<AsQhJdKc>) returns [Card<As>, Card<Qh>, Card<Jd>, Card<Kc>]",
        () {
      expect(
        List.from(CardSet({
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        })),
        equals([
          Card(rank: Rank.ace, suit: Suit.spade),
          Card(rank: Rank.queen, suit: Suit.heart),
          Card(rank: Rank.jack, suit: Suit.diamond),
          Card(rank: Rank.king, suit: Suit.club),
        ]),
      );
    });

    test("List.from(CardSet.empty) returns []", () {
      expect(
        List.from(CardSet.empty),
        equals([]),
      );
    });
  });
}
